"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.complete = exports.getLastToken = exports.Identifier = exports.UTILITY_ICON = exports.ALIAS_ICON = exports.FUNCTION_ICON = exports.TABLE_ICON = exports.COLUMN_ICON = exports.KEYWORD_ICON = void 0;
const sql_parser_1 = require("@joe-re/sql-parser");
const log4js_1 = __importDefault(require("log4js"));
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const logger = log4js_1.default.getLogger();
exports.KEYWORD_ICON = vscode_languageserver_types_1.CompletionItemKind.Text;
exports.COLUMN_ICON = vscode_languageserver_types_1.CompletionItemKind.Interface;
exports.TABLE_ICON = vscode_languageserver_types_1.CompletionItemKind.Field;
exports.FUNCTION_ICON = vscode_languageserver_types_1.CompletionItemKind.Property;
exports.ALIAS_ICON = vscode_languageserver_types_1.CompletionItemKind.Variable;
exports.UTILITY_ICON = vscode_languageserver_types_1.CompletionItemKind.Event;
const CLAUSES = [
    'SELECT',
    'WHERE',
    'ORDER BY',
    'GROUP BY',
    'LIMIT',
    '--',
    '/*',
    '('
];
const UNDESIRED_LITERAL = [
    '+',
    '-',
    '*',
    '$',
    ':',
    'COUNT',
    'AVG',
    'SUM',
    'MIN',
    'MAX',
    '`',
    '"',
    "'",
];
class Identifier {
    constructor(lastToken, identifier, detail, kind) {
        this.lastToken = lastToken;
        this.identifier = identifier;
        this.detail = detail || '';
        this.kind = kind;
    }
    matchesLastToken() {
        if (this.identifier.startsWith(this.lastToken)) {
            // prevent suggesting the lastToken itself, there is nothing to complete in that case
            if (this.identifier !== this.lastToken) {
                return true;
            }
        }
        return false;
    }
    toCompletionItem() {
        const idx = this.lastToken.lastIndexOf('.');
        const label = this.identifier.substr(idx + 1);
        let kindName;
        let tableAlias = '';
        if (this.kind === exports.TABLE_ICON) {
            let tableName = label;
            const i = tableName.lastIndexOf('.');
            if (i > 0) {
                tableName = label.substr(i + 1);
            }
            tableAlias = makeTableAlias(tableName);
            kindName = 'table';
        }
        else {
            kindName = 'column';
        }
        const item = {
            label: label,
            detail: `${kindName} ${this.detail}`,
            kind: this.kind,
        };
        if (this.kind == exports.TABLE_ICON) {
            item.insertText = `${label} AS ${tableAlias}`;
        }
        return item;
    }
}
exports.Identifier = Identifier;
// Gets the last token from the given string considering that tokens can contain dots.
function getLastToken(sql) {
    const match = sql.match(/^(?:.|\s)*[^A-z0-9\.:'](.*?)$/);
    if (match) {
        let prevToken = '';
        let currentToken = match[1];
        while (currentToken != prevToken) {
            prevToken = currentToken;
            currentToken = prevToken.replace(/\[.*?\]/, '');
        }
        return currentToken;
    }
    return sql;
}
exports.getLastToken = getLastToken;
function getFromNodesFromClause(sql) {
    try {
        return (0, sql_parser_1.parseFromClause)(sql);
    }
    catch (_e) {
        // no-op
        return null;
    }
}
function makeTableAlias(tableName) {
    if (tableName.length > 3) {
        return tableName.substr(0, 3);
    }
    return tableName;
}
function makeTableName(table) {
    if (table.catalog) {
        return table.catalog + '.' + table.database + '.' + table.tableName;
    }
    else if (table.database) {
        return table.database + '.' + table.tableName;
    }
    return table.tableName;
}
class Completer {
    constructor(schema, sql, pos, jupyterLabMode) {
        this.lastToken = '';
        this.candidates = [];
        this.error = '';
        this.isSpaceTriggerCharacter = false;
        this.isDotTriggerCharacter = false;
        this.schema = schema;
        this.sql = sql;
        this.pos = pos;
        this.jupyterLabMode = jupyterLabMode;
    }
    complete() {
        var _a;
        const target = this.getRidOfAfterCursorString();
        logger.debug(`target: ${target}`);
        this.lastToken = getLastToken(target);
        const idx = this.lastToken.lastIndexOf('.');
        this.isSpaceTriggerCharacter = this.lastToken === '';
        this.isDotTriggerCharacter = !this.isSpaceTriggerCharacter && (idx == (this.lastToken.length - 1));
        try {
            const ast = (0, sql_parser_1.parse)(target);
            this.addCandidatesForParsedStatement(ast);
        }
        catch (e) {
            logger.debug('error');
            logger.debug(e);
            if (e.name !== 'SyntaxError') {
                throw e;
            }
            const parsedFromClause = getFromNodesFromClause(this.sql);
            if (parsedFromClause) {
                const fromNodes = this.getAllNestedFromNodes(((_a = parsedFromClause === null || parsedFromClause === void 0 ? void 0 : parsedFromClause.from) === null || _a === void 0 ? void 0 : _a.tables) || []);
                const fromNodeOnCursor = this.getFromNodeByPos(fromNodes);
                if (fromNodeOnCursor && fromNodeOnCursor.type === 'incomplete_subquery') {
                    // Incomplete sub query 'SELECT sub FROM (SELECT e. FROM employees e) sub'
                    this.addCandidatesForIncompleteSubquery(fromNodeOnCursor);
                }
                else {
                    this.addCandidatesForSelectQuery(e, fromNodes);
                    this.addCandidatesForJoins(e.expected || [], fromNodes);
                }
            }
            else if (e.message === 'EXPECTED COLUMN NAME') {
                this.addCandidatesForInsert();
            }
            else {
                this.addCandidatesForError(e);
            }
            this.error = { label: e.name, detail: e.message, line: e.line, offset: e.offset };
        }
        return this.candidates;
    }
    toCompletionItemForFunction(f) {
        const item = {
            label: f.name,
            detail: 'function',
            kind: exports.FUNCTION_ICON,
            documentation: f.description,
        };
        return item;
    }
    toCompletionItemForAlias(alias) {
        const item = {
            label: alias,
            detail: 'alias',
            kind: exports.ALIAS_ICON,
        };
        return item;
    }
    toCompletionItemForKeyword(name) {
        const item = {
            label: name,
            kind: exports.KEYWORD_ICON,
            detail: 'keyword',
        };
        return item;
    }
    addCandidatesForBasicKeyword() {
        CLAUSES
            .map(v => this.toCompletionItemForKeyword(v))
            .forEach(v => this.addCandidateIfStartsWithLastToken(v));
    }
    // Check if parser expects us to terminate a single quote value or double quoted column name
    // SELECT TABLE1.COLUMN1 FROM TABLE1 WHERE TABLE1.COLUMN1 = "hoge.
    // We don't offer the ', the ", the ` as suggestions
    addCandidatesForExpectedLiterals(expected) {
        const literals = expected.filter(v => v.type === 'literal').map(v => v.text);
        const uniqueLiterals = [...new Set(literals)];
        uniqueLiterals
            .filter(v => !UNDESIRED_LITERAL.includes(v))
            .map(v => {
            switch (v) {
                case 'ORDER':
                    return 'ORDER BY';
                case 'GROUP':
                    return 'GROUP BY';
                case 'LEFT':
                    return 'LEFT JOIN';
                case 'RIGHT':
                    return 'RIGHT JOIN';
                case 'INNER':
                    return 'INNER JOIN';
                default:
                    return v;
            }
        })
            .map(v => this.toCompletionItemForKeyword(v))
            .forEach(v => this.addCandidateIfStartsWithLastToken(v));
    }
    getColumnRefByPos(columns) {
        return columns.find(v => 
        // guard against ColumnRefNode that don't have a location,
        // for example sql functions that are not known to the parser
        v.location &&
            (v.location.start.line === this.pos.line + 1 && v.location.start.column <= this.pos.column) &&
            (v.location.end.line === this.pos.line + 1 && v.location.end.column >= this.pos.column));
    }
    isPosInLocation(location) {
        return (location.start.line === this.pos.line + 1 && location.start.column <= this.pos.column) &&
            (location.end.line === this.pos.line + 1 && location.end.column >= this.pos.column);
    }
    /**
     * Finds the most deeply nested FROM node that have a range encompasing the position.
     * In cases such as SELECT * FROM T1 JOIN (SELECT * FROM (SELECT * FROM T2 <pos>))
     * We will get a list of nodes like this
     * SELECT * FROM T1
     * (SELECT * FROM
     *    (SELECT * FROM T2))
     * The idea is to reverse the list so that the most nested queries come first. Then
     * apply a filter to keep only the FROM nodes which encompass the position and take
     * the first one from that resulting list.
     * @param fromNodes
     * @param pos
     * @returns
     */
    getFromNodeByPos(fromNodes) {
        return fromNodes
            .reverse()
            .filter(tableNode => this.isPosInLocation(tableNode.location))
            .shift();
    }
    /**
     * Given a table returns all possible ways to refer to it.
     * That is by table name only, using the database scope,
     * using the catalog and database scopes.
     * @param table
     * @returns
     */
    allTableNameCombinations(table) {
        const names = [table.tableName];
        if (table.database)
            names.push(table.database + '.' + table.tableName);
        if (table.catalog)
            names.push(table.catalog + '.' + table.database + '.' + table.tableName);
        return names;
    }
    addCandidatesForTables(tables) {
        tables
            .flatMap(table => this.allTableNameCombinations(table))
            .map(aTableNameVariant => {
            return new Identifier(this.lastToken, aTableNameVariant, '', exports.TABLE_ICON);
        })
            .filter(item => item.matchesLastToken())
            .map(item => item.toCompletionItem())
            .forEach(item => this.addCandidate(item));
    }
    addCandidatesForColumnsOfAnyTable(tables) {
        tables
            .flatMap(table => table.columns)
            .map(column => {
            return new Identifier(this.lastToken, column.columnName, column.description, exports.COLUMN_ICON);
        })
            .filter(item => item.matchesLastToken())
            .map(item => item.toCompletionItem())
            .forEach(item => this.addCandidate(item));
    }
    addCandidate(item) {
        // JupyterLab requires the dot or space character preceeding the <tab> key pressed
        // If the dot or space character are not added to the label then searching
        // in the list of suggestion does not work.
        // Here we fix this issue by adding the dot or space character
        // to the filterText and insertText.
        // TODO: report this issue to JupyterLab-LSP project. 
        if (this.jupyterLabMode) {
            const text = item.insertText || item.label;
            if (this.isSpaceTriggerCharacter) {
                item.insertText = ' ' + text;
                item.filterText = ' ' + text;
            }
            else if (this.isDotTriggerCharacter) {
                item.insertText = '.' + text;
                item.filterText = '.' + text;
            }
        }
        this.candidates.push(item);
    }
    addCandidateIfStartsWithLastToken(item) {
        if (item.label.startsWith(this.lastToken)) {
            this.addCandidate(item);
        }
    }
    addCandidatesForIncompleteSubquery(incompleteSubquery) {
        const parsedFromClause = getFromNodesFromClause(incompleteSubquery.text);
        try {
            (0, sql_parser_1.parse)(incompleteSubquery.text);
        }
        catch (e) {
            if (e.name !== 'SyntaxError') {
                throw e;
            }
            const fromText = incompleteSubquery.text;
            const newPos = parsedFromClause ? {
                line: this.pos.line - (incompleteSubquery.location.start.line - 1),
                column: this.pos.column - incompleteSubquery.location.start.column + 1
            } : { line: 0, column: 0 };
            const completer = new Completer(this.schema, fromText, newPos, this.jupyterLabMode);
            completer.complete().forEach(item => this.addCandidate(item));
        }
    }
    createTablesFromFromNodes(fromNodes) {
        return fromNodes.reduce((p, c) => {
            if (c.type !== 'subquery') {
                return p;
            }
            if (!Array.isArray(c.subquery.columns)) {
                return p;
            }
            const columns = c.subquery.columns.map(v => {
                if (typeof v === 'string') {
                    return null;
                }
                return { columnName: v.as || (v.expr.type === 'column_ref' && v.expr.column) || '', description: 'alias' };
            });
            return p.concat({ database: null, columns, tableName: c.as });
        }, []);
    }
    /**
     * INSERT INTO TABLE1 (C
     */
    addCandidatesForInsert() {
        this.addCandidatesForColumnsOfAnyTable(this.schema.tables);
    }
    addCandidatesForError(e) {
        this.addCandidatesForExpectedLiterals(e.expected || []);
        this.addCandidatesForFunctions();
        this.addCandidatesForTables(this.schema.tables);
    }
    addCandidatesForSelectQuery(e, fromNodes) {
        const subqueryTables = this.createTablesFromFromNodes(fromNodes);
        const schemaAndSubqueries = this.schema.tables.concat(subqueryTables);
        this.addCandidatesForSelectStar(fromNodes, schemaAndSubqueries);
        this.addCandidatesForExpectedLiterals(e.expected || []);
        this.addCandidatesForFunctions();
        this.addCandidatesForScopedColumns(fromNodes, schemaAndSubqueries);
        this.addCandidatesForAliases(fromNodes);
        this.addCandidatesForTables(schemaAndSubqueries);
        if (logger.isDebugEnabled())
            logger.debug(`candidates for error returns: ${JSON.stringify(this.candidates)}`);
    }
    addCandidatesForJoins(expected, fromNodes) {
        let joinType = '';
        if ('INNER'.startsWith(this.lastToken))
            joinType = 'INNER';
        if ('LEFT'.startsWith(this.lastToken))
            joinType = 'LEFT';
        if ('RIGH'.startsWith(this.lastToken))
            joinType = 'RIGHT';
        if (joinType && (expected.map(v => v.text).find(v => v === 'JOIN'))) {
            if (fromNodes && fromNodes.length > 0) {
                const fromNode = fromNodes[0];
                const fromAlias = fromNode.as || fromNode.table;
                const fromTable = this.schema.tables.find(table => this.tableMatch(fromNode, table));
                this.schema.tables
                    .filter(table => table != fromTable)
                    .forEach(table => {
                    table.columns
                        .filter(column => fromTable === null || fromTable === void 0 ? void 0 : fromTable.columns.map(col => col.columnName).includes(column.columnName))
                        .map(column => {
                        return {
                            tableName: makeTableName(table),
                            alias: makeTableAlias(table.tableName),
                            columnName: column.columnName
                        };
                    })
                        .map(match => {
                        const label = `${joinType} JOIN ${match.tableName} AS ${match.alias} ON ${match.alias}.${match.columnName} = ${fromAlias}.${match.columnName}`;
                        return {
                            label: label,
                            detail: 'utility',
                            kind: exports.UTILITY_ICON,
                        };
                    })
                        .forEach(item => this.addCandidate(item));
                });
            }
        }
    }
    getRidOfAfterCursorString() {
        return this.sql.split('\n').filter((_v, idx) => this.pos.line >= idx).map((v, idx) => idx === this.pos.line ? v.slice(0, this.pos.column) : v).join('\n');
    }
    addCandidatesForParsedDeleteStatement(ast) {
        if (this.isPosInLocation(ast.table.location)) {
            this.addCandidatesForTables(this.schema.tables);
        }
        else if (ast.where && this.isPosInLocation(ast.where.expression.location)) {
            const expr = ast.where.expression;
            if (expr.type === 'column_ref') {
                this.addCandidatesForColumnsOfAnyTable(this.schema.tables);
            }
        }
    }
    completeSelectStatement(ast) {
        if (Array.isArray(ast.columns)) {
            this.addCandidateIfStartsWithLastToken(this.toCompletionItemForKeyword('FROM'));
            this.addCandidateIfStartsWithLastToken(this.toCompletionItemForKeyword('AS'));
        }
    }
    /**
     * Recursively pull out the FROM nodes (including sub-queries)
     * @param tableNodes
     * @returns
     */
    getAllNestedFromNodes(tableNodes) {
        return tableNodes.flatMap(tableNode => {
            var _a;
            let result = [tableNode];
            if (tableNode.type == 'subquery') {
                const subTableNodes = ((_a = tableNode.subquery.from) === null || _a === void 0 ? void 0 : _a.tables) || [];
                result = result.concat(this.getAllNestedFromNodes(subTableNodes));
            }
            return result;
        });
    }
    addCandidatesForParsedSelectQuery(ast) {
        var _a;
        this.addCandidatesForBasicKeyword();
        this.completeSelectStatement(ast);
        if (!ast.distinct) {
            this.addCandidateIfStartsWithLastToken(this.toCompletionItemForKeyword('DISTINCT'));
        }
        const columnRef = this.findColumnAtPosition(ast);
        if (!columnRef) {
            this.addJoinCondidates(ast);
        }
        else {
            const parsedFromClause = getFromNodesFromClause(this.sql);
            const fromNodes = ((_a = parsedFromClause === null || parsedFromClause === void 0 ? void 0 : parsedFromClause.from) === null || _a === void 0 ? void 0 : _a.tables) || [];
            const subqueryTables = this.createTablesFromFromNodes(fromNodes);
            const schemaAndSubqueries = this.schema.tables.concat(subqueryTables);
            if (columnRef.table) {
                // We know what table/alias this column belongs to
                // Find the corresponding table and suggest it's columns
                this.addCandidatesForScopedColumns(fromNodes, schemaAndSubqueries);
            }
            else {
                // Column is not scoped to a table/alias yet
                // Could be an alias, a talbe or a function
                this.addCandidatesForAliases(fromNodes);
                this.addCandidatesForTables(schemaAndSubqueries);
                this.addCandidatesForFunctions();
            }
        }
        if (logger.isDebugEnabled())
            logger.debug(`parse query returns: ${JSON.stringify(this.candidates)}`);
    }
    addCandidatesForParsedStatement(ast) {
        if (logger.isDebugEnabled())
            logger.debug(`getting candidates for parse query ast: ${JSON.stringify(ast)}`);
        if (!ast.type) {
            this.addCandidatesForBasicKeyword();
        }
        else if (ast.type === 'delete') {
            this.addCandidatesForParsedDeleteStatement(ast);
        }
        else if (ast.type === 'select') {
            this.addCandidatesForParsedSelectQuery(ast);
        }
        else {
            console.log(`AST type not supported yet: ${ast.type}`);
        }
    }
    addJoinCondidates(ast) {
        var _a, _b;
        // from clause: complete 'ON' keyword on 'INNER JOIN'
        if (ast.type === 'select' && Array.isArray((_a = ast.from) === null || _a === void 0 ? void 0 : _a.tables)) {
            const fromTable = this.getFromNodeByPos(((_b = ast.from) === null || _b === void 0 ? void 0 : _b.tables) || []);
            if (fromTable && fromTable.type === 'table') {
                this.addCandidatesForTables(this.schema.tables);
                this.addCandidateIfStartsWithLastToken(this.toCompletionItemForKeyword('INNER JOIN'));
                this.addCandidateIfStartsWithLastToken(this.toCompletionItemForKeyword('LEFT JOIN'));
                if (fromTable.join && !fromTable.on) {
                    this.addCandidateIfStartsWithLastToken(this.toCompletionItemForKeyword('ON'));
                }
            }
        }
    }
    findColumnAtPosition(ast) {
        var _a;
        const columns = ast.columns;
        if (Array.isArray(columns)) {
            // columns in select clause
            const columnRefs = columns.map((col) => col.expr).filter((expr) => !!expr);
            if (ast.type === 'select' && ((_a = ast.where) === null || _a === void 0 ? void 0 : _a.expression)) {
                // columns in where clause  
                columnRefs.push(ast.where.expression);
            }
            // column at position
            const columnRef = this.getColumnRefByPos(columnRefs);
            if (logger.isDebugEnabled())
                logger.debug(JSON.stringify(columnRef));
            return columnRef;
        }
        return undefined;
    }
    addCandidatesForFunctions() {
        console.time('addCandidatesForFunctions');
        if (!this.lastToken) {
            // Nothing was typed, return all lowercase functions
            this.schema.functions
                .map(func => this.toCompletionItemForFunction(func))
                .forEach(item => this.addCandidate(item));
        }
        else {
            // If user typed the start of the function
            const lower = this.lastToken.toLowerCase();
            const isTypedUpper = (this.lastToken != lower);
            this.schema.functions
                // Search using lowercase prefix
                .filter(v => v.name.startsWith(lower))
                // If typed string is in upper case, then return upper case suggestions
                .map(v => {
                if (isTypedUpper)
                    v.name = v.name.toUpperCase();
                return v;
            })
                .map(v => this.toCompletionItemForFunction(v))
                .forEach(item => this.addCandidate(item));
        }
        console.timeEnd('addCandidatesForFunctions');
    }
    makeColumnName(alias, columnName) {
        return alias ? alias + '.' + columnName : columnName;
    }
    addCandidatesForSelectStar(fromNodes, tables) {
        console.time('addCandidatesForSelectStar');
        tables.flatMap(table => {
            return fromNodes.filter((fromNode) => this.tableMatch(fromNode, table))
                .map((fromNode) => fromNode.as || fromNode.table)
                .filter(_alias => this.lastToken.toUpperCase() === 'SELECT' // complete SELECT keyword
                || this.lastToken === '') // complete at space after SELECT
                .map(alias => {
                const columnNames = table.columns
                    .map(col => this.makeColumnName(alias, col.columnName))
                    .join(',\n');
                const label = `Select all columns from ${alias}`;
                let prefix = '';
                if (this.lastToken) {
                    prefix = this.lastToken + '\n';
                }
                return {
                    label: label,
                    insertText: prefix + columnNames,
                    filterText: prefix + label,
                    detail: 'utility',
                    kind: exports.UTILITY_ICON,
                };
            });
        }).forEach(item => this.addCandidate(item));
        console.timeEnd('addCandidatesForSelectStar');
    }
    addCandidatesForScopedColumns(fromNodes, tables) {
        console.time('addCandidatesForScopedColumns');
        tables.flatMap(table => {
            return fromNodes.filter((fromNode) => this.tableMatch(fromNode, table))
                .map((fromNode) => fromNode.as || fromNode.table)
                .filter(alias => this.lastToken.startsWith(alias + '.'))
                .flatMap(alias => table.columns.map(col => {
                return new Identifier(this.lastToken, this.makeColumnName(alias, col.columnName), col.description, exports.COLUMN_ICON);
            }));
        })
            .filter(item => item.matchesLastToken())
            .map(item => item.toCompletionItem())
            .forEach(item => this.addCandidate(item));
        console.timeEnd('addCandidatesForScopedColumns');
    }
    /**
     * Test if the given table matches the fromNode.
     * @param fromNode
     * @param table
     * @returns
     */
    tableMatch(fromNode, table) {
        // Assume table matches from node and disprove
        let matchingTable = true;
        if (fromNode.type == 'subquery') {
            // If we have an alias it should match the subquery table name
            if (fromNode.as && fromNode.as != table.tableName) {
                matchingTable = false;
            }
        }
        // Regular tables
        // If we have a from node with a table name
        // and it does not match the table, we know it's not a match 
        else {
            if (fromNode.table && fromNode.table != table.tableName) {
                matchingTable = false;
            }
            else if (fromNode.db && fromNode.db != table.database) {
                matchingTable = false;
            }
            else if (fromNode.catalog && fromNode.catalog != table.catalog) {
                matchingTable = false;
            }
        }
        return matchingTable;
    }
    addCandidatesForAliases(fromNodes) {
        fromNodes
            .map((fromNode) => fromNode.as)
            .filter(aliasName => aliasName && aliasName.startsWith(this.lastToken))
            .map(aliasName => this.toCompletionItemForAlias(aliasName))
            .forEach(item => this.addCandidate(item));
    }
}
function complete(sql, pos, schema = { tables: [], functions: [] }, jupyterLabMode = false) {
    console.time('complete');
    if (logger.isDebugEnabled())
        logger.debug(`complete: ${sql}, ${JSON.stringify(pos)}`);
    const completer = new Completer(schema, sql, pos, jupyterLabMode);
    const candidates = completer.complete();
    console.timeEnd('complete');
    return { candidates: candidates, error: completer.error };
}
exports.complete = complete;
//# sourceMappingURL=complete.js.map